using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Threading.Tasks;
using TeachR.ApplicationManager;
using UnityEngine;

namespace TeachR.WebCommunication
{
    /// <summary>
    /// Main interface to the coach for all of the application logic.
    /// </summary>
    public static class CoachRtc
    {
        public const int SocketPort = 10000;
        public const string SocketPath = "SockServer";

        /// <summary>
        /// Listener for persistent events, i.e. events that don't need to be cleaned up at scene change
        /// </summary>
        public static RtcEvents Persistent { get; } = new();

        private static readonly List<Msg> InboundStack = new();

        public static bool Active { get; private set; }

        /// <summary>
        /// Listener for transient events, i.e. events that are specific to a current scene and that are reset after
        /// the scene ends.
        /// </summary>
        public static RtcEvents Transient { get; } = new();

        [Obsolete]
        public static Task Init()
        {
            SocketServer.HostServer(IPAddress.Any, SocketPort, SocketPath);
            SocketServer.OnIncomingMessage += HandleMessage;

            SocketServer.OnRtcOpen += () => {
                Active = true;
                CheckFunctionAvailability();
                ControllerManager.instance.CheckControllerStatus();
            };
            SocketServer.OnRtcClose += () => Active = false;

            Application.quitting += SocketServer.CloseServer;

            WebsocketDelegateHooks.OnFixedUpdate += RunEventLoop;
            RegisterFunctionCheckResultHandler(HandleFunctionCheckResult);
            RegisterScriptInfoListHandler(HandleScriptInfoListRequest);
            RegisterScriptInfoHandler(HandleScriptInfoUpdate);
            RegisterDataCollectionHandler(HandleDataCollectionParameter);
            EventLogger.Instance.OnScriptInfoListChanged += HandleScriptInfoListEntryChanged;

            return Task.CompletedTask;
        }

        /// <summary>
        /// This function is used to request from Unity whether the web app provides certain functions.
        /// </summary>
        private static void CheckFunctionAvailability()
        {
            List<string> functionsToCheck = new List<string> {
                "behave",
                "behavesfdsfdsfsfdsfd", // test-value, expected to be false
                "bootstrap",
                "disturbanceScores",
                "highlight",
                "replayState",
                "replaySync",
                "setImpulseButtonText",
                "submitRecording",
                "syncSlide",
                "syncTeacher"};
            var payload = new { functions = functionsToCheck };
            Send("checkFunctions", payload);
        }

        private static void RegisterFunctionCheckResultHandler(Action<JsonMsg> handler)
        {
            Persistent.On("functionCheckResult", handler);
        }        
        
        private static void RegisterScriptInfoListHandler(Action<JsonMsg> handler)
        {
            Persistent.On("requestScriptInfoList", handler);
        }

        private static void RegisterScriptInfoHandler(Action<JsonMsg> handler)
        {
            Persistent.On("manageScriptInfo", handler);
        }

        private static void RegisterDataCollectionHandler(Action<JsonMsg> handler)
        {
            Persistent.On("dataCollectionParameters", handler);
        }

        private static void HandleFunctionCheckResult(JsonMsg msg)
        {
            Debug.Log($"[Debug] Function check result received. DoTo: Do something with this information. Data: {msg.content}");
        }

        private static void HandleScriptInfoListRequest(JsonMsg msg)
        {
            var infoList = EventLogger.Instance.GetScriptInfoListAsJson();
            Send("sendScriptInfoList", infoList);
            EventLogger.Instance.AddEventListenerAfterConnection();
        }

        private static void HandleDataCollectionParameter(JsonMsg msg)
        {
            var data = msg.GetValue<Dictionary<string, object>>();

            string name = data["logfileName"].ToString();
            bool selectAll = Convert.ToBoolean(data["autoselectAll"]);

            EventLogger.Instance.measurementName = name;
            EventLogger.Instance.autoSelectAllScripts = selectAll;
        }


        private static void HandleScriptInfoUpdate(JsonMsg msg)
        {
            var data = msg.GetValue<Dictionary<string, object>>();

            string scriptAlias = data["scriptAlias"].ToString();
            bool logValue = Convert.ToBoolean(data["logValue"]);

            foreach (var scriptInfo in EventLogger.Instance.scriptInfoList)
            {
                if (scriptInfo.scriptAlias == scriptAlias)
                    scriptInfo.SetLogValueByWebsite(logValue);
            }
        }

        private static void HandleScriptInfoListEntryChanged(ScriptInfo scriptInfo)
        {
            var info = scriptInfo.GetScriptInfoAsJson();
            Send("updateScriptInfoEntry", info);
        }

        /// <summary>
        /// This function checks whether certain controller scripts are available in Unity. 
        /// If not, the corresponding functionalities should be deactivated on the Coach website.
        /// </summary>
        private static void CheckControllerAvailabilityInScene()
        {
            List<string> controllerToCheck = new List<string> {
                "ImpulseController",
                "BehaviourController",
                "ScenarioController",
                "SpeechToolsPLACEHOLDER",
                "LabController",
                "ToAddMoreScripts"
            };

            var allGameObjects = UnityEngine.SceneManagement.SceneManager.GetActiveScene()
                                    .GetRootGameObjects()
                                    .SelectMany(go => go.GetComponentsInChildren<Transform>(true))
                                    .Select(t => t.gameObject);

            var missingControllers = controllerToCheck.Where(controller => !allGameObjects.Any(go => go.GetComponent(controller) != null)).ToList();

            if (missingControllers.Any())
            {
                Debug.Log("The following controllers were not found in the scene: " + string.Join(", ", missingControllers));
            }
        }

        /// <summary>
        /// Converts parameters into a valid message to send to the frontend. The messages are probably 
        /// sent from website-control\src\features\behaviourControls\behaviourGroups\BehaviourManagement\SendBehaviour.jsx .
        /// </summary>
        /// <param name="type">message type</param>
        /// <param name="payload">object payload (will be stringified)</param>
        public static void Send(string type, object payload)
        {
            SocketServer.Send(type, payload);
        }

        /// <summary>
        /// Ensures that the callback function is executed only once the network is ready.
        /// </summary>
        /// <param name="callback">function to invoke</param>
        public static void RegisterOnReady(Action callback)
        {
            if (Active) callback();
            else SocketServer.OnRtcOpen += callback;
        }

        /// <summary>
        /// Handles messages from Server to Unity.
        /// </summary>
        /// <param name="arg1">Keywords like themeChange</param>
        /// <param name="arg2">Json string to add context to arg1, e.g. {"theme":"Bunsenbrenner","skip":true,"globalKnowledge":false}</param>
        private static void HandleMessage(string arg1, string arg2)
        {
            InboundStack.Add(new Msg { type = arg1, payload = arg2 });
            //Debug.Log($"[DEBUG] handle incomming message: {arg1} payload: {arg2}");
            if (arg1.Equals("configurationParameters"))
            {
                Debug.Log($"[ToDo] handle incomming message: {arg1} payload: {arg2}. These parameters currently have no effect.");
                var json = JObject.Parse(arg2);
                MenuDataHolder.StudentCount = (int)json["numStudents"];
                MenuDataHolder.LevelOfMisbehavior = (int)json["disorderLevel"];
                MenuDataHolder.IsPresentation = ((string)json["startingSituation"]).Equals("seatwork");
                MenuDataHolder.IsAutomaticIntervention = !((string)json["interventionControl"]).Equals("manually");
            }
            else if (arg1.Equals("speechtoolParameters"))
            {
                Debug.Log($"[ToDo] received speechtool parameters: {arg2}. This parameter currently has no effect.");
                // ToDo: Use parameter (enableSpeechtools [bool])
            }
        }

        private static void RunEventLoop()
        {
            foreach (Msg msg in InboundStack)
            {
                Persistent.HandleMessage(msg.type, msg.payload);
                Transient.HandleMessage(msg.type, msg.payload);
            }

            InboundStack.Clear();
        }
    }

    internal class Msg
    {
        // ReSharper disable once InconsistentNaming (JSON data)
        public string type;

        // ReSharper disable once InconsistentNaming (JSON data)
        public string payload;
    }
}