using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using TeachR.ApplicationManager;
using UnityEngine;

namespace TeachR.WebCommunication
{
    /// <summary>
    /// This class is intended to provide game objects for (controller) scripts. 
    /// It is used to check which components of the website should be available 
    /// and which should not.
    /// </summary>
    public class ControllerManager : MonoBehaviour
    {
        [SerializeField] private GameObject impulseController;
        //[SerializeField] private GameObject behaviourController;
        [SerializeField] private GameObject scenarioController;
        [SerializeField] private GameObject speechToolsController;
        [SerializeField] private GameObject labController;
        [SerializeField] private GameObject ToAddMoreScriptsController;

        //[SerializeField] private List<GameObject> controllerGameObjects = new ();
        [SerializeField] private List<ControllerScriptEntry> controllerGameObjects = new();

        public static ControllerManager instance;

        private void Awake()
        {
            // Singleton
            if (instance != null && instance != this)
                Destroy(this);
            else
            {
                instance = this;
                InitGlobals.WebsocketInitHook += Init;
            }
        }


        private Task Init()
        {
            controllerGameObjects.Add(new ControllerScriptEntry("ImpulseController", impulseController));
            controllerGameObjects.Add(new ControllerScriptEntry("ScenarioController", scenarioController));
            controllerGameObjects.Add(new ControllerScriptEntry("SpeechToolsController", speechToolsController));
            controllerGameObjects.Add(new ControllerScriptEntry("LabController", labController));

            RegisterAvailableControllersHandler(CheckControllerStatus);

            return Task.CompletedTask;
        }

        private void RegisterAvailableControllersHandler(Action<JsonMsg> handler)
        {
            CoachRtc.Persistent.On("askForControllerAvailability", handler);
        }


        private void SendControllerStatus()
        {
            // send the information to the website (as json)
            var payload = new
            {
                controllers = controllerGameObjects.ConvertAll(controller => new
                {
                    scriptName = controller.scriptName,
                    available = controller.available
                })
            };
            CoachRtc.Send("receiveControllerAvailability", payload);
        }

        /// <summary>
        /// Updates the information about the controllers and then sends it to the website.
        /// </summary>
        /// <param name="message">Message is always null</param>
        public void CheckControllerStatus(JsonMsg message = null)
        {
            foreach (var controller in controllerGameObjects)
            {
                // set the availability to true if the game object is provided, activated and contains the desired script.
                controller.SetAvailability();
            }

            SendControllerStatus();
        }
    }

    /// <summary>
    /// Helper class to combine the controller scripts you are looking for, 
    /// the associated GameObjects in the scene and the availability status 
    /// in one structure.
    /// </summary>
    [System.Serializable]
    public struct ControllerScriptEntry
    {
        public string scriptName;
        public GameObject gameObject;
        public bool available;

        public ControllerScriptEntry(string script, GameObject container, bool available = false)
        {
            scriptName = script;
            gameObject = container;
            gameObject = container;
            this.available = available;
        }

        /// <summary>
        /// Checks whether a gameObject is specified, whether it is activated in the scene and,
        /// if so, whether it has a script with the specified type. 
        /// </summary>
        /// <returns>Returns true if the GameObject <see cref="gameObject"/> has 
        /// a script of type <see cref="scriptName"/>, otherwise false.</returns>
        public void SetAvailability()
        {
            available = false;

            // gameobject exists
            if (gameObject)
            {
                // gameobject enabled
                if (gameObject.activeSelf)
                {
                    if (Type.GetType(scriptName) != null)
                    {
                        if (gameObject.GetComponent(scriptName))
                        {
                            available = true;
                        }
                        else
                            Debug.Log($"Will disable {scriptName}-functionailty on coach-website");
                    }
                    else
                        Debug.LogWarning($"No script with type {scriptName} known. Check for typo or maybe the namespace is not known.");
                }
            }
        }
    }
}
