using System;
using System.Collections.Generic;
using UnityEngine;

namespace TeachR.WebCommunication
{
    /// <summary>
    /// Event listener, that allows to register many string-identified event names.
    /// </summary>
    public class RtcEvents
    {
        private readonly Dictionary<string, Action<JsonMsg>> _actions = new();

        /// <summary>
        /// Gets an iterable of all registered event keys.
        /// </summary>
        public IEnumerable<string> Registered => _actions.Keys;

        /// <summary>
        /// Gets number of registered delegates for the given event key.
        /// </summary>
        /// <param name="key">key for lookup</param>
        /// <returns>number of invoked delegates</returns>
        public int Count(string key) => _actions[key].GetInvocationList().Length;

        /// <summary>
        /// Adds a new callback to the event of the specified name.
        /// </summary>
        /// <param name="msgType">name of the event to listen to</param>
        /// <param name="callback">callback to use when the event occurs</param>
        public void On(string msgType, Action<JsonMsg> callback)
        {
            if (_actions.TryGetValue(msgType, out _))
            {
                _actions[msgType] += callback;
            }
            else _actions[msgType] = callback;
        }

        /// <summary>
        /// Removes the specified callback from the event.
        /// </summary>
        /// <param name="msgType">name of the event to remove callback from</param>
        /// <param name="callback">callback to remove</param>
        public void Off(string msgType, Action<JsonMsg> callback)
        {
            if (_actions.TryGetValue(msgType, out _))
            {
                _actions[msgType] -= callback;
            }
            else
            {
                Debug.LogWarning($"Tried to remove non-existing event of type {msgType}");
            }
        }

        internal void HandleMessage(string msgType, string payload)
        {
            JsonMsg json = new JsonMsg()
            {
                content = payload
            };

            if (_actions.TryGetValue(msgType, out Action<JsonMsg> callback)) callback(json);
        }

        /// <summary>
        /// Resets the Event listeners.
        /// </summary>
        internal void Reset() => _actions.Clear();
    }
}